(************************************************************************)
(* This program is free software; you can redistribute it and/or        *)
(* modify it under the terms of the GNU General Public License as       *)
(* published by the Free Software Foundation; either version 2 of the   *)
(* License, or (at your option) any later version.                      *)
(*                                                                      *)
(* This program is distributed in the hope that it will be useful, but  *)
(* WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    *)
(* General Public License for more details.                             *)
(*                                                                      *)
(* Written and (c) by Jean-Baptiste Rouquier                            *)
(************************************************************************)
(* ordre : vertical, diagonale montante, diagonale descendante *)
open Utils;;

type rangee =
  | Souhait of int (* int\in[1..9] reprsente la couleur *)
  | Fixe of int*int (* couleur, score courant. Ligne complte ou non. *)
  | Mort (* Deux couleurs diffrentes : cette rangee ne donnera pas de point. *)
type configuration =
    {rangees: rangee array array;
     mutable coups_possibles: (int*int) list}
type piece = int array
  (* verticale, diagonale montante, diagonale descendante.
     piece.(0)\in[1,5,9], piece.(1)\in[2,6,7], piece.(2)\in[3,4,8]*)
type position = int*int
let rangs_of_pos (x,y) = [|x; y-x+2; y|];; (* toujours verticale, montante, descendante *)

module Evaluation :sig
  type t = private {
    mutable minus_nb_morts: int; mutable minus_valeur_morts: int;
      (* toujours ngatif, pour permettre la comparaison gnrique "<" *)
(*     mutable nb_completes: int; mutable valeur_completes: int; *)
    mutable nb_prolonges: int; mutable valeur_prolonges: int;
    mutable nb_souhaits_honnores: int; mutable valeur_souhaits: int;}
  val gt : t -> t -> bool
  val main : configuration -> piece -> position -> t
end = struct
  type t = {
    mutable minus_nb_morts: int; mutable minus_valeur_morts: int;
(*     mutable nb_completes: int; mutable valeur_completes: int; *)
    mutable nb_prolonges: int; mutable valeur_prolonges: int;
    mutable nb_souhaits_honnores: int; mutable valeur_souhaits: int;}
  let empty_evaluation () = {
    minus_nb_morts=0; minus_valeur_morts=0;
(*     nb_completes=0; valeur_completes=0; *)
    nb_prolonges=0; valeur_prolonges=0;
    nb_souhaits_honnores=0; valeur_souhaits=0;}
  let gt = (>)

  let main config piece pos =
    let rangs = rangs_of_pos pos in
    let result = empty_evaluation () in
    for i=0 to 2 do
      match config.rangees.(i).(rangs.(i)) with
        | Souhait couleur ->
            if couleur = piece.(i) then (
              result.nb_souhaits_honnores <- result.nb_souhaits_honnores +1;
              result.valeur_souhaits <- result.valeur_souhaits +couleur;)
        | Fixe (couleur,score) ->
            if couleur = piece.(i) then (
              result.nb_prolonges <- result.nb_prolonges +1;
              result.valeur_prolonges <- result.valeur_prolonges +couleur;
            ) else (
              result.minus_nb_morts <- result.minus_nb_morts -1;
              result.minus_valeur_morts <- result.minus_valeur_morts -couleur
            )
        | Mort -> ()
    done;
    result;;
end

let init () =
  let coups_possibles = ref [] in
  for i=4 downto 0 do
    for j=4 downto 0 do
      if not (List.mem (i,j) [0,3; 0,4; 1,4; 3,0; 4,0; 4,1])
      then coups_possibles:= (i,j) :: !coups_possibles
    done;
  done;
  {coups_possibles = !coups_possibles;
  rangees = [|
    [|Souhait 1; Souhait 9; Souhait 5; Souhait 9; Souhait 1|];
    [|Souhait 2; Souhait 7; Souhait 6; Souhait 7; Souhait 2|];
    [|Souhait 3; Souhait 8; Souhait 4; Souhait 8; Souhait 3|];
    (* autre essai, qui ne change pas significativement le score moyen, seulement augmente l'cart type de 2
       (avez-vous une meilleure ide?) : *)
(*     [|Souhait 1; Souhait 5; Souhait 9; Souhait 5; Souhait 1|]; *)
(*     [|Souhait 2; Souhait 6; Souhait 7; Souhait 6; Souhait 2|]; *)
(*     [|Souhait 3; Souhait 4; Souhait 8; Souhait 4; Souhait 3|]; *)
  |]}

let best_pos config piece =
  List.map (fun pos -> pos, Evaluation.main config piece pos) config.coups_possibles
  |> list_max (fun (_,eval1) (_,eval2) -> Evaluation.gt eval1 eval2)
  |> fst

let place config piece = (* appelle best_pos et modifie config *)
  let pos = best_pos config piece in
  config.coups_possibles <- List.filter ((<>) pos) config.coups_possibles;
  let rangs = rangs_of_pos pos in
  for i=0 to 2 do
    let rangee = config.rangees.(i) in
    let rang = rangs.(i) in
    match rangee.(rang) with
      | Souhait _ -> rangee.(rang) <- Fixe (piece.(i),piece.(i))
      | Fixe (couleur,score) ->
          if couleur = piece.(i)
          then rangee.(rang) <- Fixe (couleur,score+couleur)
          else rangee.(rang) <- Mort
      | Mort -> ()
  done;
  pos;;

let score config =
  let result = ref 0 in
  for i=0 to 2 do
    for j=0 to 4 do
    match config.rangees.(i).(j) with
      | Souhait _ -> failwith "score"
      | Fixe (_,score) -> result += score
      | Mort -> ()
    done;
  done;
  !result;;

(*************** interface ***************)
(* dessin *)
(* #load "graphics.cma";; *)
open Graphics;;

let hexagone = [|0,180; 35,120; 105,120; 140,180; 105,240; 35,240|];;
(* let vertical   = [|55,120; 85,120; 85,240; 55,240|];; *)
(* let diago_mont = [|130, 197; 115, 223; 10, 162; 25, 137|];; *)
(* let diago_desc = [|115, 137; 130, 162; 25, 223; 10, 197|];; *)
(* On rordonne pour que text_pos_of_rectangle marche *)
let bandes = [|
  [| 85,240;  55,240;  55,120;  85,120;  |]; (* vertical *)
  [| 10,162;  25,137;  130,197; 115,223; |]; (* diago_mont *)
  [| 115,137; 130,162; 25,223;  10,197;  |]; (* diago_desc *)
|]

(* set_text_size 50;; *)
let text_pos_of_bande b =
  let x0,y0 = b.(0) in
  let x1,y1 = b.(1) in
  let x2,y2 = b.(2) in
  x0 + (x1-x0)/2 + (x2-x1)/ 8(*  - 4 *),
(*     - fst (text_size "9") /2, *)
  y0 + (y1-y0)/2 + (y2-y1)/ 8(*  - 6 *)
let text_positions = Array.map text_pos_of_bande bandes

let translation (a,b) (x,y) =
  x+ 105*a, y+ 120*b - 60*a;;

let couleurs =
  [|9868950; 16754175; 13715370; 4503743; 3053781; 16384094; 9622083; 16747059; 14605902|];;

let draw pos piece =
  set_color black; fill_poly (Array.map (translation pos) hexagone);
  for i=0 to 2 do
    set_color couleurs.(piece.(i)-1); fill_poly (Array.map (translation pos) bandes.(i));
    set_color black;
    let x,y = translation pos text_positions.(i) in
    let tx,ty = text_size "9" in
    moveto (x-tx/2) (y-ty/2);
    draw_string (string_of_int piece.(i));
  done;;

(* texte *)
exception Piece_inexistante
let read_piece () =
  let continue = ref true in
  let result = [|None; None; None|] in
  while !continue do
    try
      for i=0 to 2 do result.(i) <- None; done;
      print_string "piece : ";
      let n = int_of_string (read_line ()) in
      let couleurs = [n/100; (n mod 100)/10; n mod 10] in
      List.iter (fun i -> if not (1 <= i && i <= 9) then raise Piece_inexistante) couleurs;
      List.iter
        (fun c ->
           if      List.mem c [1;5;9] then result.(0) <- Some c
           else if List.mem c [2;6;7] then result.(1) <- Some c
           else if List.mem c [3;4;8] then result.(2) <- Some c)
        couleurs;
      Array.iter (function None -> raise Piece_inexistante | Some _ -> ()) result;
      continue := false;
    with
      |Failure "int_of_string" -> Printf.printf "mauvais format, entrez un entier a trois chiffres\n%!"
      |Piece_inexistante -> Printf.printf "cette piece n'existe pas\n%!";
  done;
  Array.map (function Some c -> c | None -> failwith "read_piece") result

let open_graph_tight () = open_graph " 569x628"

let joue_partie () =
  let config = init () in
  open_graph_tight ();
  Printf.printf "Entrez les trois chiffres de la piece (par exemple \"924\"), puis enter :\n%!";
  for i=1 to 19 do
    let piece = read_piece () in
    let pos = place config piece in
    draw pos piece;
  done;
  Printf.printf "score : %d.\n%!" (score config);
  ignore (read_line ());
  close_graph ()
;;

let pieces = [|
  [|1;2;3|]; [|1;2;4|]; [|1;2;8|];  [|1;6;3|]; [|1;6;4|]; [|1;6;8|];  [|1;7;3|]; [|1;7;4|]; [|1;7;8|];
  [|5;2;3|]; [|5;2;4|]; [|5;2;8|];  [|5;6;3|]; [|5;6;4|]; [|5;6;8|];  [|5;7;3|]; [|5;7;4|]; [|5;7;8|];
  [|9;2;3|]; [|9;2;4|]; [|9;2;8|];  [|9;6;3|]; [|9;6;4|]; [|9;6;8|];  [|9;7;3|]; [|9;7;4|]; [|9;7;8|];|]

open Big_int

let mean_score nb_parties =
  let total_score = ref zero_big_int in
  let somme_x_carre = ref zero_big_int in
  let ordre = Array.copy pieces in
  for partie = 1 to nb_parties do
    let config = init() in
    shuffle ordre;
    for tour = 0 to 18 do
      let piece = ordre.(tour) in
      let pos = place config piece in
      (*     draw pos piece; *)
      ()
    done;
    let score = score config in
    total_score := add_int_big_int score !total_score;
    somme_x_carre := add_int_big_int (score*score) !somme_x_carre;
  done;
  let mean = float_of_big_int !total_score /. float nb_parties in
  Printf.printf "score moyen : %f.\n%!" mean;
  Printf.printf "ecart type : %f.\n%!"
    (sqrt (float_of_big_int !somme_x_carre /. float nb_parties -. mean*.mean));
;;

let demo () =
  open_graph_tight ();
  let config = init () in
  let ordre = Array.copy pieces in
  shuffle ordre;
  for tour = 0 to 18 do
    let piece = ordre.(tour) in
    let pos = place config piece in
    draw pos piece;
    Printf.printf "piece: %i%i%i\n%!" piece.(0) piece.(1) piece.(2);
  done;
  Printf.printf "score : %d.\n%!" (score config);
  ignore (read_key ());
  close_graph ();
;;

let main = ref joue_partie;;

Arg.parse
  ["-joue", Arg.Unit (fun () -> main := joue_partie),
     "joue une partie selon les pieces entrees au clavier. C'est le comportement par dfaut.";
   "-demo", Arg.Unit (fun () -> main := (fun () -> Random.self_init (); demo ())),
     "joue une partie en tirant au sort les pieces et affiche la position finale et le score";
   "-stats", Arg.Int (fun i -> main := (fun () -> Random.self_init (); mean_score i)),
     "prend un nombre de partie a jouer (sans affichage) et calcule le score moyen. \
      Pour 10 000 000 parties, j'ai obtenu un score moyen de 132,8, avec un ecart type de 35,4."]
  (fun s -> Printf.printf "\"%s\" non pris en compte\n%!" s)
  "Automate pour Take It Easy. Une boite du jeu est necessaire pour jouer contre l'ordinateur. \
   Vous tirez les pieces et les lui indiquez, \
   il les place a l'ecran et indique son score a la fin de la manche."
;;

!main ();;
